@@ -1,10 +1,25 @@
 import inspect
 import pygame
-from simpn.simulator import SimToken, SimVar, SimEvent
+from simpn.simulator import SimToken, SimVar, SimEvent, SimProblem, Describable
+from simpn.simulator import SimTokenValue
 import simpn.visualisation as vis
+import math
+import re
 
+# matplotlib set1 colormap
+TASK_TOKEN_SHOW_COLOURS = [
+    pygame.Color("#E41A1C"),
+    pygame.Color("#377EB8"),
+    pygame.Color("#4DAF4A"),
+    pygame.Color("#984EA3"),
+    pygame.Color("#FF7F00"),
+    pygame.Color("#FFFF33"),
+    pygame.Color("#A65628"),
+    pygame.Color("#F781BF"),
+    pygame.Color("#999999"),
+]
 
-class Prototype:
+class Prototype(Describable):
     """
     Superclass for all prototypes. Contains the basic structure of a prototype, which is a composition of SimVar and SimEvent.
     A prototype must subclass this class.
@@ -58,6 +73,126 @@ class Prototype:
         self.visualization_of_edges = edges
 
 
+class TaskTokenShower(vis.Node):
+    """
+    A helper drawer that captures the in-flights tokens that are being 
+    processed by a BPMN prototype. The shower creates a group of tokens 
+    above the task, colouring them by the prefix of the tokens to highlight
+    what tokens are being processed in the groups.
+    """
+
+    def __init__(self, model_node):
+        super().__init__(model_node)
+        self._task = None 
+        self._token_radius = 3
+        self._token_dia = self._token_radius * 2
+        self._seen_token_types = {}
+        self._next_tok_colour = 0
+        self._max_rows = 3
+
+    def set_task(self, task) -> 'TaskTokenShower':
+        self._task = task 
+        return self 
+    
+    def set_pos(self, pos) -> 'TaskTokenShower':
+        self._pos = pos 
+        return self 
+    
+    def set_rows(self, rows:int) -> 'TaskTokenShower':
+        self._max_rows = rows 
+        return self
+    
+    def set_time(self, clock:float) -> 'TaskTokenShower':
+        self._curr_time = clock
+        return self 
+    
+    def get_token_colour(self, token):
+        # handle the token
+        if isinstance(token, tuple):
+            name = token[0]
+        else:
+            if not isinstance(token, str):
+                name = str(token)
+            else:
+                name = token
+        # it would be nice if we did not have to do this to find the general
+        # family that the token comes from
+        cut = re.compile("[0-9]").search(name).span()[0]
+        name = name[:cut]
+        # find a colour
+        if name not in self._seen_token_types:
+            colour = TASK_TOKEN_SHOW_COLOURS[self._next_tok_colour]
+            self._next_tok_colour += 1 
+            if (self._next_tok_colour >= len(TASK_TOKEN_SHOW_COLOURS)):
+                self._next_tok_colour = 0 
+            self._seen_token_types[name] = colour 
+        else:
+            colour = self._seen_token_types[name]
+        return colour
+    
+    def draw(self, screen:pygame.Surface):
+        """
+        Draws the groups of tokens being processed above the current task
+        """
+        if (self._task is not None):
+            tokens = []
+            node = self._task._model_node
+            if hasattr(node, '_busyvar'):
+                tokens = node._busyvar.marking
+            elif hasattr(node, '_marking'):
+                tokens = node._marking
+            else:
+                raise ValueError(f"Could not identify the in-flight tokens for the given task :: {self._task=}")
+
+            # start from the top left of the node
+            group_height = self._token_dia * 2
+            curr_x = self._pos[0] - self._task._half_width 
+            padding_x = self._token_radius
+            curr_y = self._pos[1] \
+                - self._task._half_height - group_height - self._token_radius
+            padding_y = self._token_radius
+            start_x = curr_x
+            row_end_x = start_x + self._task._width 
+            rows = 1
+
+            # loop through token groups
+            for group in tokens:
+                group_width = self._token_dia * (len(group.value) + 1)
+                
+                group_rect = pygame.Rect(
+                    curr_x, curr_y,
+                    group_width, group_height
+                )
+
+                pygame.draw.rect(
+                    screen, vis.TUE_LIGHTBLUE, group_rect, vis.LINE_WIDTH,
+                    int(group_width * 0.25)
+                )
+
+                tok_x = curr_x + self._token_radius
+                tok_y = curr_y + self._token_radius
+                # loop through values of the group
+                for tok in group.value:
+                    tok_circle = pygame.draw.circle(
+                        screen, self.get_token_colour(tok),
+                        (tok_x + self._token_radius, 
+                         tok_y + self._token_radius),
+                        self._token_radius,
+                    )
+                    tok_x += self._token_dia
+
+                # handle moving to a new row
+                curr_x += padding_x + group_width
+                if (curr_x >= row_end_x):
+                    curr_x = start_x
+                    curr_y -= padding_y + group_height
+                    rows += 1 
+                
+                # to prevent drawing a million groups
+                if (rows >= self._max_rows):
+                    break
+
+
 class BPMNStartEvent(Prototype):  
     """
     Generates a composition of SimVar and SimEvent that represents a BPMN start event.
@@ -70,11 +205,12 @@ class BPMNStartEvent(Prototype):
     :param outgoing: a list with a single SimVar in which the cases will be placed.
     :param name: the name of the start event.
     :param interarrival_time: the interarrival time with which events are generated. Can be a numeric value or a function that produces a numeric value, such as a sampling function from a random distribution.
-    :param behavior: an optional behavior describing how case_data is produced.
+    :param behavior: an optional function describing how case_data is produced.
     """
 
     def __init__(self, model, incoming, outgoing, name, interarrival_time, behavior=None):
         super().__init__(model, incoming, outgoing, name)
+        model.set_binding_priority(SimProblem.PRIORITY_QUEUE_BINDING)  # We process tokens in BPMN models FCFS
 
         if len(incoming) != 0:
             raise TypeError("Start event " + name + ": cannot have any incoming.")
@@ -95,14 +231,21 @@ class BPMNStartEvent(Prototype):
         else:
             if not callable(behavior):
                 raise TypeError("Start event " + name + ": the behavior must be a function. (Maybe you made it a function call, exclude the brackets.)")
-            if len(inspect.signature(behavior).parameters) != 0:
-                raise TypeError("Start event " + name + ": the behavior function must not have many parameters.")
-            result = model.add_event([invar], [invar, outgoing[0]], lambda a: [SimToken(name + str(int(a[len(name):]) + 1), delay=interarrival_time_f()), SimToken((a, behavior()[0].value))], name=name + "<start_event>")
+            if len(inspect.signature(behavior).parameters) > 1:
+                raise TypeError("Start event " + name + ": the behavior function must take either a single parameter, the identity or none.")
+            elif len(inspect.signature(behavior).parameters) == 0:
+                result = model.add_event([invar], [invar, outgoing[0]], lambda a: [SimToken(name + str(int(a[len(name):]) + 1), delay=interarrival_time_f()), SimToken((a, behavior()[0].value))], name=name + "<start_event>")
+            elif len(inspect.signature(behavior).parameters) == 1:
+                result = model.add_event([invar], [invar, outgoing[0]], lambda a: [SimToken(name + str(int(a[len(name):]) + 1), delay=interarrival_time_f()), behavior(a)], name=name + "<start_event>")
             self.add_event(result)
         invar.put(name + "0")
 
         model.add_prototype(self)
 
+    def get_description(self):
+        description = [(super().get_id() + ": BPMNStartEvent", Describable.Style.HEADING)]
+        return description
+
     class BPMNStartEventViz(vis.Node):
         def __init__(self, model_node):
             super().__init__(model_node)
@@ -111,7 +254,7 @@ class BPMNStartEvent(Prototype):
             pygame.draw.circle(screen, vis.TUE_LIGHTBLUE, (self._pos[0], self._pos[1]), self._width/2)
             pygame.draw.circle(screen, vis.TUE_BLUE, (self._pos[0], self._pos[1]), self._width/2, vis.LINE_WIDTH)    
             font = pygame.font.SysFont('Calibri', vis.TEXT_SIZE)
-
+            
             # draw label
             label = font.render(self._model_node.get_id(), True, vis.TUE_BLUE)
             text_x_pos = self._pos[0] - int(label.get_width()/2)
@@ -143,6 +286,7 @@ class BPMNTask(Prototype):
 
     def __init__(self, model, incoming, outgoing, name, behavior, guard=None, outgoing_behavior=None):
         super().__init__(model, incoming, outgoing, name)
+        model.set_binding_priority(SimProblem.PRIORITY_QUEUE_BINDING)  # We process tokens in BPMN models FCFS
 
         if len(incoming) < 2:
             raise TypeError("Task event " + name + ": must have at least two input parameters; the first for cases and the second for resources.")
@@ -178,6 +322,14 @@ class BPMNTask(Prototype):
 
         model.add_prototype(self)
 
+    def get_description(self):
+        description = [(super().get_id() + ": BPMNTask", Describable.Style.HEADING)]
+        description.append( (" ", Describable.Style.NORMAL) )
+        description.append( ("Marking:", Describable.Style.NORMAL) )
+        for token in self._busyvar.marking:
+            description.append( (str(token), Describable.Style.BOXED) )
+        return description
+
     class BPMNTaskViz(vis.Node):
         def __init__(self, model_node):
             super().__init__(model_node)
@@ -185,6 +337,7 @@ class BPMNTask(Prototype):
             self._height = vis.STANDARD_NODE_HEIGHT
             self._half_width =  self._width / 2
             self._half_height = self._height / 2
+            self._token_shower = TaskTokenShower(None)
 
         def draw(self, screen):
             x_pos, y_pos = int(self._pos[0] - self._width/2), int(self._pos[1] - self._height/2)
@@ -199,20 +352,11 @@ class BPMNTask(Prototype):
             text_y_pos = int((self._height - label.get_height())/2) + y_pos
             screen.blit(label, (text_x_pos, text_y_pos))
 
-            # draw marking
-            mstr = "["
-            ti = 0
-            for token in self._model_node._busyvar.marking:
-                mstr += str(token.value) + "@" + str(round(token.time, 2))
-                if ti < len(self._model_node._busyvar.marking) - 1:
-                    mstr += ", "
-                ti += 1
-            mstr += "]"
-            label = bold_font.render(mstr, True, vis.TUE_RED)
-            text_x_pos = self._pos[0] - int(label.get_width()/2)
-            text_y_pos = self._pos[1] + self._half_height + vis.LINE_WIDTH
-            screen.blit(label, (text_x_pos, text_y_pos))        
-
+            # draw tokens above the task being processed
+            self._token_shower \
+                .set_task(self) \
+                .set_pos(self.get_pos()) \
+                .draw(screen)
 
     def get_visualisation(self):
         return self.BPMNTaskViz(self)
@@ -233,6 +377,7 @@ class BPMNIntermediateEvent(Prototype):
 
     def __init__(self, model, incoming, outgoing, name, behavior, guard=None):
         super().__init__(model, incoming, outgoing, name)
+        model.set_binding_priority(SimProblem.PRIORITY_QUEUE_BINDING)  # We process tokens in BPMN models FCFS
 
         if len(incoming) < 1:
             raise TypeError("Event " + name + ": must have at least one input parameter for cases.")
@@ -253,6 +398,10 @@ class BPMNIntermediateEvent(Prototype):
 
         model.add_prototype(self)
 
+    def get_description(self):
+        description = [(super().get_id() + ": BPMNIntermediateEvent", Describable.Style.HEADING)]
+        return description
+
     class BPMNIntermediateEventViz(vis.Node):
         def __init__(self, model_node):
             super().__init__(model_node)
@@ -285,17 +434,34 @@ class BPMNEndEvent(Prototype):
 
     def __init__(self, model, incoming, outgoing, name):
         super().__init__(model, incoming, outgoing, name)
+        model.set_binding_priority(SimProblem.PRIORITY_QUEUE_BINDING)  # We process tokens in BPMN models FCFS
 
         if len(incoming) != 1:
             raise TypeError("Event " + name + ": must have one input parameter for cases.")
         if len(outgoing) != 0:
             raise TypeError("Event " + name + ": must not have output parameters.")
 
-        result = model.add_event(incoming, outgoing, lambda c: [], name=name + "<end_event>")
+        self._marking = []
+        result = model.add_event(incoming, outgoing, lambda c: self.behaviour(c), name=name + "<end_event>")
         self.add_event(result)
 
         model.add_prototype(self)
 
+    def behaviour(self, c):
+        """
+        wrapper to count tokens that made it to the end event.
+        """
+        self._marking.append(SimToken(c))
+        return []
+
+    def get_description(self):
+        description = [(super().get_id() + ": BPMNEndEvent", Describable.Style.HEADING)]
+        description.append( (" ", Describable.Style.NORMAL) )
+        description.append( ("Marking:", Describable.Style.NORMAL) )
+        for token in self._marking:
+            description.append( (str(token), Describable.Style.BOXED) )
+        return description
+    
     class BPMNEndEventViz(vis.Node):
         def __init__(self, model_node):
             super().__init__(model_node)
@@ -305,6 +471,13 @@ class BPMNEndEvent(Prototype):
             pygame.draw.circle(screen, vis.TUE_BLUE, (self._pos[0], self._pos[1]), self._width/2, vis.LINE_WIDTH*2)
             font = pygame.font.SysFont('Calibri', vis.TEXT_SIZE)
 
+            # draw tokens
+            vis.TokenShower(self._model_node._marking) \
+                .set_pos(self._pos) \
+                .show_token_count(True) \
+                .set_time(self._curr_time) \
+                .draw(screen)
+
             # draw label
             label = font.render(self._model_node.get_id(), True, vis.TUE_BLUE)
             text_x_pos = self._pos[0] - int(label.get_width()/2)
@@ -319,10 +492,17 @@ class BPMNFlow(SimVar):
     """
     A SimVar that represents a BPMN Flow.
     It is just a SimVar with a different visualisation.
+    The flow can have a priority, which makes sense if the flow is the single flow into a BPMN task,
+    the priority then determines the order in which cases are processed by the task. Default is FCFS.
+
+    :param model: the SimProblem to which the event composition must be added.
+    :param _id: the id of the flow.
+    :param priority: the priority with which tokens on the flow are processed, defaults to FCFS.
     """
 
-    def __init__(self, model, _id):
-        super().__init__(_id)
+    def __init__(self, model, _id, priority=None):
+        super().__init__(_id, priority=priority)
+        model.set_binding_priority(SimProblem.PRIORITY_QUEUE_BINDING)  # We process tokens in BPMN models FCFS
 
         model.add_prototype_var(self)
 
@@ -339,21 +519,12 @@ class BPMNFlow(SimVar):
             x, y = self._pos
             pygame.draw.circle(screen, vis.TUE_BLUE, (x, y), self._width)
 
-            bold_font = pygame.font.SysFont('Calibri', vis.TEXT_SIZE, bold=True)
-            
-            # draw marking
-            mstr = "["
-            ti = 0
-            for token in self._model_node.marking:
-                mstr += str(token.value) + "@" + str(round(token.time, 2))
-                if ti < len(self._model_node.marking) - 1:
-                    mstr += ", "
-                ti += 1
-            mstr += "]"
-            label = bold_font.render(mstr, True, vis.TUE_RED)
-            text_x_pos = self._pos[0] - int(label.get_width()/2)
-            text_y_pos = self._pos[1] + self._half_height + vis.LINE_WIDTH + int(label.get_height())
-            screen.blit(label, (text_x_pos, text_y_pos))        
+            # draw tokens 
+            vis.TokenShower(self._model_node.marking) \
+                .show_token_count() \
+                .set_pos(self._pos) \
+                .set_time(self._curr_time) \
+                .draw(screen)    
 
     def get_visualisation(self):
         return self.BPMNFlowViz(self)
@@ -377,6 +548,7 @@ class BPMNExclusiveSplitGateway(Prototype):
 
     def __init__(self, model, incoming, outgoing, label, behavior):
         super().__init__(model, incoming, outgoing, label)
+        model.set_binding_priority(SimProblem.PRIORITY_QUEUE_BINDING)  # We process tokens in BPMN models FCFS
 
         if len(incoming) != 1:
             raise TypeError("Gateway " + label + ": must have at exactly one input parameter for cases.")
@@ -406,6 +578,10 @@ class BPMNExclusiveSplitGateway(Prototype):
 
         model.add_prototype(self)
 
+    def get_description(self):
+        description = [(super().get_id() + ": BPMNExclusiveSplitGateway", Describable.Style.HEADING)]
+        return description
+
     class BPMNExclusiveSplitGatewayViz(vis.Node):
         def __init__(self, model_node):
             super().__init__(model_node)
@@ -444,6 +620,7 @@ class BPMNExclusiveJoinGateway(Prototype):
 
     def __init__(self, model, incoming, outgoing, label):
         super().__init__(model, incoming, outgoing, label)
+        model.set_binding_priority(SimProblem.PRIORITY_QUEUE_BINDING)  # We process tokens in BPMN models FCFS
 
         if len(incoming) < 2:
             raise TypeError("Gateway " + label + ": must have multiple input parameter for cases.")
@@ -463,6 +640,10 @@ class BPMNExclusiveJoinGateway(Prototype):
 
         model.add_prototype(self)
 
+    def get_description(self):
+        description = [(super().get_id() + ": BPMNExclusiveJoinGateway", Describable.Style.HEADING)]
+        return description
+    
     class BPMNExclusiveJoinGatewayViz(vis.Node):
         def __init__(self, model_node):
             super().__init__(model_node)
@@ -513,6 +694,7 @@ class BPMNParallelSplitGateway(SimEvent):
 
     def __init__(self, model, incoming, outgoing, label):
         super().__init__(label)
+        model.set_binding_priority(SimProblem.PRIORITY_QUEUE_BINDING)  # We process tokens in BPMN models FCFS
 
         if len(incoming) != 1:
             raise TypeError("Gateway " + label + ": must have at exactly one input parameter for cases.")
@@ -526,6 +708,10 @@ class BPMNParallelSplitGateway(SimEvent):
 
         model.add_prototype_event(self)
 
+    def get_description(self):
+        description = [(super().get_id() + ": BPMNParallelSplitGateway", Describable.Style.HEADING)]
+        return description
+
     class BPMNParallelSplitGatewayViz(vis.Node):
         def __init__(self, model_node):
             super().__init__(model_node)
@@ -561,6 +747,7 @@ class BPMNParallelJoinGateway(SimEvent):
 
     def __init__(self, model, incoming, outgoing, label, behavior=None):
         super().__init__(label)
+        model.set_binding_priority(SimProblem.PRIORITY_QUEUE_BINDING)  # We process tokens in BPMN models FCFS
 
         if len(incoming) < 2:
             raise TypeError("Gateway " + label + ": must have at least two input parameter for cases.")
@@ -580,6 +767,10 @@ class BPMNParallelJoinGateway(SimEvent):
 
         model.add_prototype_event(self)
 
+    def get_description(self):
+        description = [(super().get_id() + ": BPMNParallelJoinGateway", Describable.Style.HEADING)]
+        return description
+
     class BPMNParallelJoinGatewayViz(vis.Node):
         def __init__(self, model_node):
             super().__init__(model_node)
