@@ -2,8 +2,32 @@ import inspect
 from sortedcontainers import SortedList
 import simpn.visualisation as vis
 
+from random import choice
+from typing import List, Dict 
+from enum import Enum, auto
+from copy import deepcopy
 
-class SimVar:
+
+class Describable:
+    """
+    A class that can provide a description of itself for visualisation purposes.
+    """
+    class Style(Enum):
+        HEADING = auto()
+        NORMAL = auto()
+        BOLD = auto()
+        BOXED = auto()
+
+    def get_description(self):
+        """
+        Returns a description of the object. This is can be used for visualisation purposes.
+        The description is a list of tuples, where each tuple has the form ('text', styles) a text element with the specified styles, where styles is a list of elements of Describable.Style.
+        :return: a list of tuples.
+        """
+        return []
+
+
+class SimVar(Describable):
     """
     A simulation variable SimVar has an identifier and a marking.
     A simulation variable can have multiple values. These values are available at particular times.
@@ -16,8 +40,13 @@ class SimVar:
     :param priority: a function that takes a token as input and returns a value that is used to sort the tokens in the order in which they will be processed (lower values first). The default is processing in the order of the time of the token.
     """
 
-    def __init__(self, _id, priority=lambda token: token.time):
+    def __init__(self, _id, priority=None):
         self._id = _id
+        if priority == None:
+            priority = lambda token: token.time
+            self._sorted_by_time = True
+        else:
+            self._sorted_by_time = False
         self.marking = SortedList(key=priority)
         self.checkpoints = dict()
         self.queue = SimVarQueue(self)
@@ -89,6 +118,14 @@ class SimVar:
     def set_invisible_edges(self):
         self.visualize_edges = False
     
+    def get_description(self):
+        description = [(self._id + ": SimVar", Describable.Style.HEADING) ]        
+        description.append( (" ", Describable.Style.NORMAL) )
+        description.append( ("Marking:", Describable.Style.NORMAL) )
+        for token in self.marking:
+            description.append( (str(token), Describable.Style.BOXED) )
+        return description
+
 
 class SimVarQueue(SimVar):
     """
@@ -101,6 +138,8 @@ class SimVarQueue(SimVar):
     def __init__(self, simvar):
         self._id = simvar._id + SimVarQueue.QUEUE_SUFFIX
         self.simvar = simvar
+        self.visualize = True
+        self.visualize_edges = True
 
     @property
     def marking(self):
@@ -149,10 +188,11 @@ class SimVarTime(SimVar):
         self.checkpoints = dict()
         self.problem = problem
         self.visualize = True
+        self.visualize_edges = True
 
     @property
     def marking(self):
-        token = SimToken(self.problem.clock)
+        token = SimTokenTime(None, problem=self.problem)
         return SortedList([token])
 
     def put(self, value, time=0):
@@ -171,7 +211,7 @@ class SimVarTime(SimVar):
         return self.__str__()
 
 
-class SimEvent:
+class SimEvent(Describable):
     """
     A simulation event SimEvent that can happen when tokens are available on all of its
     incoming SimVar and its guard evaluates to True. When it happens, it consumes a token from each
@@ -273,6 +313,172 @@ class SimEvent:
 
         self.visualization_of_edges = edges
 
+    def get_description(self):
+        description = [(self._id + ": SimEvent", Describable.Style.HEADING)]
+        return description
+
+class SimTokenValue:
+    """
+    A data structure to enable the handling of values and named attributes
+    associated with tokens. By convention, the first possible value held
+    by the structure is it's identity or `id`.
+
+    :param id: 
+        :type:`object`
+        the identity of this token.
+
+    :param ...values:
+        :type:`List[object]`
+        a list of unnamed attributes, or comma passed values passed to the
+        constructor.
+
+    :param ...dict_values:
+        :type:`Dict[str,object]`
+        a dictionary of named attributes, or named parameters passed to the
+        constructor.
+
+    --------
+    Examples
+    --------
+
+    With this class, values of tokens can be built in a pythonic manner
+    without much fuss and avoids using tuples everywhere.
+
+    .. code-block:: python 
+        # create a new value!
+        val = SimTokenValue("a1", 1, 'a', mary='alice', joe={ 'value' : 5})
+        # you can grab the identity of the token
+        id = val.id # or val[0]
+        # but you can still pretend to have a tuple
+        print(val[1])
+        # but for named arguments, you can use the dot notation on the value
+        print(val.mary)
+        print(val[3]) # same as above
+        # to ensure that immutable state is kept you can quickly make an
+        # equivalent value reference
+        copy_of_val = val.clone()
+        # you can quickly update values as well
+        val.mary= 'ham'
+        val['joe'] = 'spam'
+    """
+
+    def __init__(self, id, *values:List[object], **dict_values:Dict[str,object]):
+        tmp = [id] + list(values)
+        
+        self._id = id
+        self._named_start = len(values) + 1
+        self._names = sorted(list(dict_values.keys()))
+        for key in self._names:
+            if (hasattr(self, key)):
+                raise ValueError(
+                    f"SimTokenValue has duplicate keys in parameter list" \
+                    + f" :: {key}"
+                    )
+            setattr(self, key, dict_values[key])
+            tmp.append(dict_values[key])
+
+        self._values = tmp
+
+    @property
+    def id(self) -> object:
+        """
+        Returns the identity of the token.
+        """
+        return self._id
+
+    def clone(self) -> 'SimTokenValue':
+        """
+        Returns a clone of itself. When working within a behaviour and you
+        want to make changes to the state of the token, best practice is to
+        do so on a clone.
+
+        :returns SimTokenValue:
+        """
+        return deepcopy(self)
+    
+    def __hash__(self):
+        return self._values.__hash__()
+
+    def __str__(self):
+        result = "("
+        for val in self._values[:self._named_start]:
+            result += f"{val}, "
+        for name in self._names:
+            result += f"{name}={self[name]}, "
+        result = result[:-2] + ")"
+        return result
+
+    def __repr__(self):
+        ret = f"SimTokenValue({repr(self._values[0])}, "
+
+        for val in self._values[1:self._named_start]:
+            ret += f"{repr(val)}, "
+
+        named_values = dict( 
+            (key,val) 
+            for key,val 
+            in zip(self._names, self._values[self._named_start:])
+        )
+        ret += f"**{repr(named_values)})"
+
+        return ret
+    
+    def __eq__(self, value):
+        if isinstance(value, SimTokenValue):
+            return self._values == value._values
+        else:
+            return tuple(self._values) == value
+    
+    def __getitem__(self, key, default_val=None):
+        if isinstance(key, str):
+            if hasattr(self, key):
+                return getattr(self, key)
+            else:
+                return default_val
+        elif isinstance(key, int) or isinstance(key, slice):
+            return self._values[key]
+        else:
+            raise TypeError(f"The given key {key} was neither a str nor a int.")
+        
+    def __setattr__(self, key, value):
+        if (hasattr(self, '_names') and hasattr(self, '_values') and hasattr(self, '_named_start')):
+            if (key in ["id", "_id"]):
+                raise TypeError("You cannot change the identity of a token.")
+            
+            if hasattr(self, key):
+                index = self._named_start + self._names.index(key)
+                self._values[index] = value 
+            else:
+                self._names.append(key)
+                self._values.append(value)
+        super().__setattr__(key, value)
+        
+    def __setitem__(self, key, value):
+        if key == 0:
+            raise TypeError("You cannot change the identity of a token.")
+
+        if isinstance(key, str):
+            setattr(self, key, value)
+            if hasattr(self, key):
+                index = self._named_start + self._names.index(key)
+                self._values[index] = value 
+            else:
+                self._names.append(key)
+                self._values.append(value)
+        elif isinstance(key, int):
+            self._values[key] = value 
+            if key >= self._named_start:
+                index = key - self._named_start
+                key = self._names[index]
+                setattr(self, key, value)
+        else:
+            raise TypeError(f"The given key {key} was neither a str nor a int.")
+
+    def __len__(self):
+        return len(self._values)
+    
+    def __iter__(self):
+        return iter(self._values)
 
 class SimToken:
     """
@@ -327,6 +533,29 @@ class SimToken:
         return SimToken(self.value, self.time)
 
 
+class SimTokenTime(SimToken):
+    """
+    A token that represents the simulation time.
+    This token only always has a single value, which is the simulation time.
+    The value will be taken from the clock of the simulator through an attribute function,
+    so it always has the correct value at the time of firing.
+    """
+    def __init__(self, value, time=0, delay=0, problem=None):
+        self._problem = problem
+        
+    @property
+    def value(self):
+        return self._problem.clock
+    
+    @property
+    def time(self):
+        return 0
+    
+    @property
+    def delay(self):
+        return 0
+
+
 class SimProblem:
     """
     A simulation problem SimProblem, which consists of a collection of simulation variables SimVar and a collection of simulation events SimEvent.
@@ -337,9 +566,24 @@ class SimProblem:
     it can happen on any of these values.
 
     :param debugging: if set to True, produces more information for debugging purposes (defaults to True).
-    :param binding_priority: a function that takes a list of binding as input and returns the binding that will be selected in case there are multiple possible bindings to fire.
+    :param binding_priority: a function that takes a list of binding as input and returns the binding that will be selected in case there are multiple possible bindings to fire (defaults to random selection).
     """
-    def __init__(self, debugging=True, binding_priority=lambda bindings: bindings[0]):
+    @staticmethod
+    def RANDOM_BINDING(bindings):
+        """ Random selection between all bindings. Deprecated: since we now have prioritized bindings per event, this function is equivalent to PRIORITY_QUEUE_BINDING. """
+        return choice(bindings)
+    
+    @staticmethod
+    def PRIORITY_BINDING(bindings):
+        """ Prioritized bindings per event, and prioritized selection between events. """
+        return bindings[0]
+
+    @staticmethod
+    def PRIORITY_QUEUE_BINDING(bindings):
+        """ Prioritized bindings per event, but random selection between events. """
+        return choice(bindings)
+
+    def __init__(self, debugging=True, binding_priority=None):
         self.places = []
         self.events = []
         self.prototypes = []
@@ -347,7 +591,10 @@ class SimProblem:
         self.clock = 0
         self._debugging = debugging
         self.clock_checkpoints = dict()
-        self.binding_priority = binding_priority
+        if binding_priority is None:
+            self.binding_priority = self.PRIORITY_QUEUE_BINDING
+        else:
+            self.binding_priority = binding_priority
 
     def __str__(self):
         result = ""
@@ -397,7 +644,7 @@ class SimProblem:
         result += "}"                
         return result
 
-    def add_var(self, name, priority=lambda token: token.time):
+    def add_var(self, name, priority=None):
         """
         Creates a new SimVar with the specified name as identifier. Adds the SimVar to the problem and returns it.
 
@@ -446,6 +693,28 @@ class SimProblem:
         self.events.append(event)
         self.id2node[name] = event
 
+    def event(self, name):
+        """
+        Returns the SimEvent with the given name.
+        Raises an exception if no such SimEvent exists.
+
+        :param name: the name of the SimEvent.
+        :return: the SimEvent with the given name or an Error.
+        """
+        if name in self.id2node:
+            if isinstance(self.id2node[name], SimEvent):
+                return self.id2node[name]
+            else:
+                raise TypeError(name + " is not a SimEvent.")
+        else:
+            raise LookupError("SimEvent " + name + ": does not exist.")
+    def transition(*args, **kwargs):
+        """
+        Returns the SimEvent with the given name. This function is a wrapper around event for people with a Petri-net background.
+        """
+        self = args[0]
+        return self.event(*(args[1:]), **kwargs)
+
 
     def var(self, name):
         """
@@ -593,75 +862,102 @@ class SimProblem:
         self.binding_priority = func
 
     @staticmethod
-    def tokens_combinations(event):
+    def evaluate_guard(variable_values, event):
         """
-        Creates a list of token combinations that are available to the specified event.
-        These are combinations of tokens that are on the incoming SimVar of the event.
-        For example, if a event has incoming SimVar a and b with tokens 1@0 on a and 2@0, 3@0 on b,
-        the possible combinations are [(a, 1@0), (b, 2@0)] and [(a, 1@0), (b, 3@0)]
+        Evaluates the guard condition for the given event on the given variable values.
+        
+        :param variable_values: the list of values for the parameters of the guard.
+        :param event: the event for which to evaluate the guard.
+        :return: True if the guard evaluates to True for the given binding, False otherwise.
+        """
+        enabled = True
+        try:
+            enabled = event.guard(*variable_values)
+        except Exception as e:
+            raise TypeError("Event " + event + ": guard generates exception for values " + str(variable_values) + ".") from e
+        if not isinstance(enabled, bool):
+            raise TypeError("Event " + event + ": guard does evaluate to a Boolean for values " + str(variable_values) + ".")
+        return enabled
+                        
+    @staticmethod
+    def bfs(incoming_places, combinations, max_time, event, look_beyond=False):
+        """        
+        Breadth-first search to find a combination of tokens that satisfy the guard.
+        If there is no guard, the first combination is returned.
+        Initially, look_beyond is False, meaning we only consider combinations, where each token <= max_time.
+        If no such combination is found, we set look_beyond to True and continue searching without the max_time constraint.
+        Tokens are identified by a combination of indices into the lists of tokens in the incoming places
+        (a combination is a tuple (i_1, i_2, ..., i_n) where i_j is the index of the token to be identified in the j-th incoming place).
 
-        :param event: the event to return the token combinations for.
-        :return: a list of lists, each of which is a token combination.
+        :param incoming_places: A list of incoming places to the event.
+        :param combinations: A list of tuples, each representing a combination of indices, which identify the tokens to consider as a possible binding.
+        :param event: The event for which to find a valid binding.
+        :param max_time: The maximum index value to consider when look_beyond is False.
+        :param look_beyond: A boolean indicating whether to ignore the max_time constraint.
+        :return: The first combination of indices that satisfies the guard, or None if no such combination exists.
         """
-        # create all possible combinations of incoming token values
-        bindings = [[]]
-        for place in event.incoming:
-            new_bindings = []
-            for token in place.marking:  # get set of colors in incoming place
-                for binding in bindings:
-                    new_binding = binding.copy()
-                    new_binding.append((place, token))
-                    new_bindings.append(new_binding)
-            bindings = new_bindings
-        return bindings
+        next_combinations = []
+        seen = set()
+
+        for t in combinations:
+            # if there is no guard, or the guard is satisfied
+            # THEN we found a valid combination
+            if (event.guard is None or SimProblem.evaluate_guard([event.incoming[i].marking[t[i]].value for i in range(len(t))], event)):
+                return t
+            for j in range(len(t)):
+                if t[j] + 1 >= len(incoming_places[j].marking):
+                    continue
+                # if the incoming place is sorted by time, and the next token does not satisfy the time constraint, we can skip it
+                if event.incoming[j]._sorted_by_time and not look_beyond and event.incoming[j].marking[t[j] + 1].time > max_time:
+                    continue
+                u = (*t[:j], t[j] + 1, *t[j+1:])
+                if u not in seen:
+                    next_combinations.append(u)
+                    seen.add(u)
+        if len(next_combinations) == 0:
+            if look_beyond:
+                return None
+            else:
+                return SimProblem.bfs(incoming_places, combinations, max_time, event, True)
+        return SimProblem.bfs(incoming_places, next_combinations, max_time, event, look_beyond)
 
     def event_bindings(self, event):
         """
-        Calculates the set of bindings that enables the given event.
-        Each binding is a tuple ([(place, token), (place, token), ...], time) that represents a single enabling binding.
-        A binding is
-        a possible token combination (see token_combinations), for which the event's
-        guard function evaluates to True. In case there is no guard function, any combination is also a binding.
-        The time is the time at which the latest token is available.
-        For example, if a event has incoming SimVar a and b with tokens 1@2 on a and 2@3, 3@1 on b,
-        the possible bindings are ([(a, 1@2), (b, 2@3)], 3) and ([(a, 1@2), (b, 3@1)], 2)
+        Calculates the first binding that enables the given event.
+        A binding is a tuple ([(place, token), (place, token), ...], time)
+        for which the event's guard function evaluates to True. 
+        In case there is no guard function, any combination is also a binding.
+        The time of the binding is the time at which the latest token is available.
+        Which tokens are considered first is determined by the priority functions of the SimVars.
+        The max_time is the current clock time of the simulation.        
 
-        :param event: the event for which to calculate the enabling bindings.
-        :return: list of tuples ([(place, token), (place, token), ...], time)
+        :param event: the event for which to calculate the enabling binding.
+        :return: a tuple ([(place, token), (place, token), ...], time)
         """
-        if len(event.incoming) == 0:
-            raise Exception("Though it is strictly speaking possible, we do not allow events like '" + str(self) + "' without incoming arcs.")
+        incoming = event.incoming
+        if len(incoming) == 0:
+            raise Exception("Though it is strictly speaking possible, we do not allow events like '" + str(self) + "' without incoming arcs.")        
+        for place in incoming: # if any of the incoming places is empty, there is no binding
+            if len(place.marking) == 0:
+                return None
 
-        bindings = self.tokens_combinations(event)
-
-        # a binding must have all incoming places
-        nr_incoming_places = len(event.incoming)
-        new_bindings = []
-        for binding in bindings:
-            if len(binding) == nr_incoming_places:
-                new_bindings.append(binding)
-        bindings = new_bindings
-
-        # if a event has a guard, only bindings are enabled for which the guard evaluates to True
-        result = []
-        for binding in bindings:
-            variable_values = []
-            time = None
-            for (place, token) in binding:
-                variable_values.append(token.value)
-                if time is None or token.time > time:
-                    time = token.time
-            enabled = True
-            if event.guard is not None:
-                try:
-                    enabled = event.guard(*variable_values)
-                except Exception as e:
-                    raise TypeError("Event " + event + ": guard generates exception for values " + str(variable_values) + ".") from e
-                if self._debugging and not isinstance(enabled, bool):
-                    raise TypeError("Event " + event + ": guard does evaluate to a Boolean for values " + str(variable_values) + ".")
-            if enabled:
-                result.append((binding, time))
-        return result
+        # we perform a breadth-first search over the combinations of tokens, starting with the first token of each incoming place, the first one found that satisfies the guard is returned
+        # if no combination is found, we return None
+        combinations = [tuple(0 for _ in range(len(incoming)))] # start with the first combination, which is the first token of each incoming place
+        max_time = self.clock # initially, we only consider tokens that are available at or before the current clock time
+        indices = SimProblem.bfs(incoming, combinations, max_time, event)
+        if indices is None: # we did not find a binding
+            return None
+        else: # we found a binding, the corresponding tokens are at the indices, the time is the max of the times of these tokens
+            binding = []
+            max_token_time = 0
+            for i in range(len(incoming)):
+                place = incoming[i]
+                token = place.marking[indices[i]]
+                binding.append((place, token))
+                if token.time > max_token_time:
+                    max_token_time = token.time
+            return (binding, max_token_time)
 
     def bindings(self):
         """
@@ -673,7 +969,9 @@ class SimProblem:
         timed_bindings = []
         min_enabling_time = None
         for t in self.events:
-            for (binding, time) in self.event_bindings(t):
+            timed_binding = self.event_bindings(t)
+            if timed_binding is not None:
+                (binding, time) = timed_binding
                 timed_bindings.append((binding, time, t))
                 if min_enabling_time is None or time < min_enabling_time:
                     min_enabling_time = time
@@ -681,16 +979,10 @@ class SimProblem:
         # if there are no such bindings, set the clock to the earliest time at which there are
         if min_enabling_time is not None and min_enabling_time > self.clock:
             self.clock = min_enabling_time
-            # We now also need to update the bindings, because the SimVarTime may have changed and needs to be updated.
-            # TODO This is inefficient, because we are recalculating all bindings, while we only need to recalculate the ones that have SimVarTime in their inflow.
-            timed_bindings = [] 
-            for t in self.events:
-                for (binding, time) in self.event_bindings(t):
-                    timed_bindings.append((binding, time, t))
 
         # now return the untimed bindings + the timed bindings that have time <= clock
-        return [(binding, time, t) for (binding, time, t) in timed_bindings if time <= self.clock]
-
+        return [(binding, self.clock, t) for (binding, time, t) in timed_bindings if time <= self.clock]
+    
     def fire(self, timed_binding):
         """
         Fires the specified timed binding.
@@ -779,31 +1071,4 @@ class SimProblem:
                     else:
                         reporter.callback(timed_binding)
             else:
-                active_model = False
-
-
-def event(sim_problem: SimProblem, inflow: list, outflow: list, guard=None):
-    """
-    A decorator that can be used to turn a Python function into a SimEvent.
-    The event will be added to the specified sim_problem.
-    If they do not yet exist, SimVar will be added for each parameter of the Python function and for each element in outflow.
-    The parameters of the function will be treated as inflow variables of the SimEvent (by name).
-    The elements of outflow must be names of SimVar and will be treated as outflow variables of the SimEvent (by name).
-    The name of the SimEvent will be the name of the Python function.
-
-    :param sim_problem: the SimProblem to which to add the SimEvent.
-    :param inflow: a list of names of inflow variables. Just there for documentation purposes, it is not being used.
-    :param outflow: a list of names of outflow variables.
-    :param guard: an optional guard for the SimEvent.
-    :return: the Python function.
-    """
-    def decorator_event(func):
-        parameters = inspect.signature(func).parameters
-        for parameter in list(parameters.keys()) + outflow:
-            if parameter not in sim_problem.id2node:
-                sim_problem.add_var(parameter)
-        inflow_svars = [sim_problem.id2node[parameter] for parameter in parameters]
-        outflow_svars = [sim_problem.id2node[parameter] for parameter in outflow]
-        sim_problem.add_event(inflow_svars, outflow_svars, func, guard=guard)
-        return func
-    return decorator_event
+                active_model = False
\ No newline at end of file
